from common.lib_support import ensure_kfserving_installed

ensure_kfserving_installed()

import json  # noqa: E402
import logging  # noqa: E402
from http import HTTPStatus  # noqa: E402

import numpy as np  # noqa: E402
import tornado.web  # noqa: E402
from common.serialization import DeepNumpyEncoder  # noqa: E402
from common.serialization import truss_msgpack_deserialize  # noqa: E402
from common.serialization import truss_msgpack_serialize  # noqa: E402
from common.util import \
    assign_request_to_inputs_instances_after_validation  # noqa: E402
from kfserving.handlers.http import HTTPHandler  # noqa: E402
from kfserving.kfserver import LivenessHandler  # noqa: E402
from kfserving.kfserver import (HealthHandler, KFServer,  # noqa: E402
                                ListHandler)
from pythonjsonlogger import jsonlogger  # noqa: E402


def _configure_logging():
    json_log_handler = logging.StreamHandler()
    json_log_handler.setFormatter(jsonlogger.JsonFormatter('%(asctime)s %(levelname)s %(message)s'))
    logger = logging.getLogger()
    for handler in logger.handlers:
        logger.removeHandler(handler)
    logger.addHandler(json_log_handler)


class TrussHTTPBinaryHandler(HTTPHandler):
    def validate(self, request):
        if ("instances" in request and not isinstance(request["instances"], (list, np.ndarray))) or \
           ("inputs" in request and not isinstance(request["inputs"], (list, np.ndarray))):
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Expected \"instances\" or \"inputs\" to be a list or NumPy ndarray"
            )
        return assign_request_to_inputs_instances_after_validation(request)


class TrussHTTPHandler(HTTPHandler):
    def validate(self, request):
        if ("instances" in request and not isinstance(request["instances"], list)) or \
           ("inputs" in request and not isinstance(request["inputs"], list)):
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Expected \"instances\" or \"inputs\" to be a list"
            )
        return assign_request_to_inputs_instances_after_validation(request)


class TrussPredictHandler(TrussHTTPBinaryHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = truss_msgpack_deserialize(self.request.body)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.predict(request)
        response = model.postprocess(response)
        try:
            final_response = truss_msgpack_serialize(response)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e
            )
        self.write(final_response)


class TrussExplainHandler(TrussHTTPBinaryHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = truss_msgpack_deserialize(self.request.body)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.explain(request)
        response = model.postprocess(response)
        try:
            final_response = truss_msgpack_serialize(response)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e
            )
        self.write(final_response)


class PredictHandler(TrussHTTPHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = json.loads(self.request.body)
        except json.decoder.JSONDecodeError as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.predict(request)
        response = model.postprocess(response)
        try:
            final_response = json.dumps(response, cls=DeepNumpyEncoder)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e
            )
        self.write(final_response)


class ExplainHandler(TrussHTTPHandler):
    def post(self, name: str):
        model = self.get_model(name)
        try:
            body = json.loads(self.request.body)
        except json.decoder.JSONDecodeError as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.BAD_REQUEST,
                reason="Unrecognized request format: %s" % e
            )
        request = self.validate(body)
        request = model.preprocess(request)
        response = model.explain(request)
        response = model.postprocess(response)
        try:
            final_response = json.dumps(response, cls=DeepNumpyEncoder)
        except Exception as e:
            raise tornado.web.HTTPError(
                status_code=HTTPStatus.INTERNAL_SERVER_ERROR,
                reason="Unable to serialize model prediction to response: %s" % e
            )
        self.write(final_response)


class TrussServer(KFServer):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        _configure_logging()

    def create_application(self):
        return tornado.web.Application([
            # Server Liveness API returns 200 if server is alive.
            (r"/", LivenessHandler),
            (r"/v1/models",
             ListHandler, dict(models=self.registered_models)),
            # Model Health API returns 200 if model is ready to serve.
            (r"/v1/models/([a-zA-Z0-9_-]+)",
             HealthHandler, dict(models=self.registered_models)),
            (r"/v1/models/([a-zA-Z0-9_-]+):predict",
             PredictHandler, dict(models=self.registered_models)),
            (r"/v1/models/([a-zA-Z0-9_-]+):predict_binary",
             TrussPredictHandler, dict(models=self.registered_models)),
            (r"/v1/models/([a-zA-Z0-9_-]+):explain",
             ExplainHandler, dict(models=self.registered_models)),
            (r"/v1/models/([a-zA-Z0-9_-]+):explain_binary",
             TrussExplainHandler, dict(models=self.registered_models)),
        ])



from contextlib import contextmanager
from pathlib import Path

from truss.local.local_config_handler import LocalConfigHandler

CONFIG_CONTENT = '''
secrets:
    secret_name: secret_value
'''


def test_get_config(tmp_path):
    with _truss_config_dir(tmp_path, CONFIG_CONTENT):
        config = LocalConfigHandler.get_config()
        assert config.secrets['secret_name'] == 'secret_value'


def test_set_secret(tmp_path):
    with _truss_config_dir(tmp_path, CONFIG_CONTENT):
        LocalConfigHandler.set_secret('another_secret_name', 'another_secret_value')
        config = LocalConfigHandler.get_config()
        assert config.secrets['secret_name'] == 'secret_value'
        assert config.secrets['another_secret_name'] == 'another_secret_value'


def test_set_secret_override(tmp_path):
    with _truss_config_dir(tmp_path, CONFIG_CONTENT):
        LocalConfigHandler.set_secret('secret_name', 'another_secret_value')
        config = LocalConfigHandler.get_config()
        assert config.secrets['secret_name'] == 'another_secret_value'


def test_remove_secret(tmp_path):
    with _truss_config_dir(tmp_path, CONFIG_CONTENT):
        LocalConfigHandler.remove_secret('secret_name')
        config = LocalConfigHandler.get_config()
        assert 'secret_name' not in config.secrets


def test_sync_secrets_mount_dir(tmp_path):
    with _truss_config_dir(tmp_path, CONFIG_CONTENT):
        LocalConfigHandler.sync_secrets_mount_dir()
        assert (tmp_path / 'secrets').exists()
        with (tmp_path / 'secrets' / 'secret_name').open() as f:
            assert f.read() == 'secret_value'


@contextmanager
def _truss_config_dir(path: Path, config_content: str):
    orig_config_dir = LocalConfigHandler.TRUSS_CONFIG_DIR
    LocalConfigHandler.TRUSS_CONFIG_DIR = path
    try:
        with (path / 'config.yaml').open('w') as f:
            f.write(config_content)
        yield
    finally:
        LocalConfigHandler.TRUSS_CONFIG_DIR = orig_config_dir



import json
from pathlib import Path
from typing import Dict, List

import yaml

from truss.constants import CONFIG_FILE
from truss.truss_config import TrussConfig
from truss.types import ModelFrameworkType


class TrussSpec:
    """Helper class for easy access to information in a Truss."""
    def __init__(self, truss_dir: Path) -> None:
        self._truss_dir = truss_dir
        self._config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)

    @property
    def truss_dir(self) -> Path:
        return self._truss_dir

    @property
    def config_path(self) -> Path:
        return self._truss_dir / CONFIG_FILE

    @property
    def data_dir(self) -> Path:
        return self._truss_dir / self._config.data_dir

    @property
    def model_module_dir(self) -> Path:
        return self._truss_dir / self._config.model_module_dir

    @property
    def model_class_filepath(self) -> Path:
        conf = self._config
        return self._truss_dir / conf.model_module_dir / conf.model_class_filename

    @property
    def config(self) -> TrussConfig:
        return self._config

    @property
    def python_version(self) -> str:
        return self._config.python_version

    @property
    def cpu(self) -> str:
        return self._config.resources.cpu

    @property
    def json_string(self) -> str:
        return json.dumps(self._config.to_dict())

    @property
    def memory(self) -> str:
        return self._config.resources.memory

    @property
    def use_gpu(self) -> str:
        return self._config.resources.use_gpu

    @property
    def model_module_name(self) -> str:
        return str(Path(self._config.model_class_filename).with_suffix(''))

    @property
    def model_module_fullname(self) -> str:
        return f'{self._config.model_module_dir}.{self.model_module_name}'

    @property
    def model_class_name(self) -> str:
        return self._config.model_class_name

    @property
    def model_framework_type(self) -> ModelFrameworkType:
        return self._config.model_framework

    @property
    def model_framework_name(self) -> str:
        return self.model_framework_type.value

    @property
    def requirements(self) -> List[str]:
        return self._config.requirements

    @property
    def requirements_txt(self) -> str:
        return _join_lines(self._config.requirements)

    @property
    def system_packages(self) -> List[str]:
        return self._config.system_packages

    @property
    def system_packages_txt(self) -> str:
        return _join_lines(self._config.system_packages)

    @property
    def environment_variables(self) -> Dict[str, str]:
        return self._config.environment_variables

    @property
    def examples_path(self) -> Path:
        return self._truss_dir / self._config.examples_filename

    @property
    def examples(self) -> Dict[str, Dict]:
        with self.examples_path.open() as yaml_file:
            return yaml.safe_load(yaml_file)

    @property
    def yaml_string(self) -> str:
        with self.config_path.open() as yaml_file:
            return yaml_file.read()

    @property
    def secrets(self) -> Dict[str, str]:
        return self._config.secrets


def _join_lines(lines: List[str]) -> str:
    return '\n'.join(lines) + '\n'



def model_supports_predict_proba(model):
    if not hasattr(model, 'predict_proba'):
        return False
    if hasattr(model, '_check_proba'):  # noqa eg Support Vector Machines *can* predict proba if they made certain choices while training
        try:
            model._check_proba()
            return True
        except AttributeError:
            return False
    return True


def assign_request_to_inputs_instances_after_validation(request):
    # we will treat "instances" and "inputs" the same
    if "instances" in request and "inputs" not in request:
        request["inputs"] = request["instances"]
    elif "inputs" in request and "instances" not in request:
        request["instances"] = request["inputs"]
    return request



from typing import Dict, List

import torch
from torch import package


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs['data_dir']
        config = kwargs['config']
        self._model_metadata = config['model_metadata']
        self._model = None
        self._model_dtype = None
        self._device = torch.device(
            "cuda:0" if torch.cuda.is_available() and config['resources']['use_gpu'] else "cpu"
        )

    def load(self):
        imp = package.PackageImporter(self._data_dir / 'model' / self._model_metadata['torch_package_file'])
        package_name = self._model_metadata['torch_model_package_name']
        model_pickle_filename = self._model_metadata['torch_model_pickle_filename']
        self._model = imp.load_pickle(package_name, model_pickle_filename)
        self._model_dtype = list(self._model.parameters())[0].dtype

    def preprocess(self, request: Dict) -> Dict:
        """
        Incorporate pre-processing required by the model if desired here.

        These might be feature transformations that are tightly coupled to the model.
        """
        return request

    def postprocess(self, request: Dict) -> Dict:
        """
        Incorporate post-processing required by the model if desired here.
        """
        return request

    def predict(self, request: Dict) -> Dict[str, List]:
        response = {}
        with torch.no_grad():
            inputs = torch.tensor(
                request["inputs"], dtype=self._model_dtype, device=self._device
            )
            response = {}
            response["predictions"] = self._model(inputs).tolist()
            return response



import os
import pathlib

SKLEARN = 'sklearn'
TENSORFLOW = 'tensorflow'
KERAS = 'keras'
PYTORCH = 'pytorch'
CUSTOM = 'custom'
HUGGINGFACE_TRANSFORMER = 'huggingface_transformer'

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
CODE_DIR = pathlib.Path(BASE_DIR, 'truss')

TEMPLATES_DIR = pathlib.Path(CODE_DIR, 'templates')
SERVER_CODE_DIR = TEMPLATES_DIR / 'server'

# Alias for TEMPLATES_DIR
SERVING_DIR = TEMPLATES_DIR

REQUIREMENTS_TXT_FILENAME = 'requirements.txt'
SERVER_REQUIREMENTS_TXT_FILENAME = 'server_requirements.txt'
SYSTEM_PACKAGES_TXT_FILENAME = 'system_packages.txt'

SERVER_DOCKERFILE_TEMPLATE_NAME = 'server.Dockerfile.jinja'
MODEL_DOCKERFILE_NAME = 'Dockerfile'

README_TEMPLATE_NAME = 'README.md.jinja'
MODEL_README_NAME = 'README.md'

CONFIG_FILE = 'config.yaml'
DOCKERFILE = "Dockerfile"
# Used to indicate whether to associate a container with Truss
TRUSS = "truss"
# Used to create unique identifier based on last time truss was updated
TRUSS_MODIFIED_TIME = "truss_modified_time"
# Path of the Truss used to identify which Truss is being referred
TRUSS_DIR = "truss_dir"



import logging
import traceback
from typing import Dict, List

import torch
from transformers import pipeline


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs['data_dir']
        config = kwargs['config']
        self._config = config
        model_metadata = config['model_metadata']
        self._transformer_config = model_metadata['transformer_config']
        self._has_named_args = model_metadata['has_named_args']
        self._has_hybrid_args = model_metadata['has_hybrid_args']
        self._model = None

    def load(self):
        transformer_config = self._transformer_config.copy()
        if torch.cuda.is_available():
            transformer_config['device'] = 0

        self._model = pipeline(
            task=self._config['model_type'],
            **transformer_config,
        )

    def preprocess(self, request: Dict) -> Dict:
        """
        Incorporate pre-processing required by the model if desired here.

        These might be feature transformations that are tightly coupled to the model.
        """
        return request

    def postprocess(self, request: Dict) -> Dict:
        """
        Incorporate post-processing required by the model if desired here.
        """
        return request

    def predict(self, request: Dict) -> Dict[str, List]:
        response = {}
        instances = request['inputs']

        with torch.no_grad():
            if self._has_named_args:
                result = [self._model(**instance) for instance in instances]
            elif self._has_hybrid_args:
                try:
                    result = []
                    for instance in instances:
                        prompt = instance.pop('prompt')
                        result.append(self._model(prompt, **instance))
                except (KeyError, AttributeError):
                    logging.error(traceback.format_exc())
                    response['error'] = {
                        'traceback':
                            f'Expected request as an object with text in "prompt"\n{traceback.format_exc()}'
                    }
                    return response
            else:
                result = self._model(instances)
        response['predictions'] = result
        return response



from jinja2 import Template

from truss.constants import README_TEMPLATE_NAME, TEMPLATES_DIR
from truss.truss_spec import TrussSpec


def generate_readme(_spec: TrussSpec) -> str:
    readme_template_path = TEMPLATES_DIR / README_TEMPLATE_NAME
    with readme_template_path.open() as readme_template_file:
        readme_template = Template(readme_template_file.read())
        # examples.yaml may not exist
        # if examples.yaml does exist, but it's empty, examples_raw is None
        examples_raw = _spec.examples if _spec.examples_path.exists() else None
        readme_contents = readme_template.render(config=_spec.config, examples=examples_raw)
    return readme_contents



import os
from pathlib import Path
from typing import Any, List

import yaml

from truss.constants import CONFIG_FILE, TEMPLATES_DIR, TRUSS
from truss.docker import kill_containers
from truss.model_frameworks import model_framework_from_model
from truss.model_inference import infer_python_version
from truss.truss_config import DEFAULT_EXAMPLES_FILENAME, TrussConfig
from truss.truss_handle import TrussHandle
from truss.types import ModelFrameworkType
from truss.utils import (build_truss_target_directory, copy_file_path,
                         copy_tree_path)


def mk_truss(
    model: Any,
    target_directory: str = None,
    data_files: List[str] = None,
    requirements_file: str = None,
) -> TrussHandle:
    """Create a Truss with the given model. A Truss is a build context designed to
    be built as a container locally or uploaded into a baseten serving environment.

    Args:
        model (an in-memory model object): A model object to be deployed (e.g. a keras, sklearn, or pytorch model
            object)
        target_directory (str, optional): The local directory target for the Truss. Otherwise a temporary directory
            will be generated
        data_files (List[str], optional): Additional files required for model operation. Can be a glob that resolves to
            files for the root directory or a directory path.
        requirements_file (str, optional): A file of packages in a PIP requirements format to be installed in the
            container environment.
    Returns:
        TrussHandle: A handle to the generated Truss that provides easy access to content inside.
    """
    model_framework = model_framework_from_model(model)
    if target_directory is None:
        target_directory_path = build_truss_target_directory(model_framework.typ().value)
    else:
        target_directory_path = Path(target_directory)
    model_framework.to_truss(model, target_directory_path)
    scaf = TrussHandle(target_directory_path)
    _update_truss_props(scaf, data_files, requirements_file)
    return scaf


def init(
    target_directory: str,
    data_files: List[str] = None,
    requirements_file: str = None,
) -> TrussHandle:
    """
    Initialize an empty placeholder Truss. A Truss is a build context designed
    to be built as a container locally or uploaded into a baseten serving
    environment. This placeholder structure can be filled to represent ML
    models.

    Args:
        target_directory: Absolute or relative path of the directory to create
                          Truss in. The directory is created if it doesn't exist.
    """
    target_directory_path = Path(target_directory)
    target_directory_path.mkdir(parents=True, exist_ok=True)
    config = TrussConfig(
        model_type='custom',
        model_framework=ModelFrameworkType.CUSTOM,
        python_version=infer_python_version(),
    )

    # Create data dir
    (target_directory_path / config.data_dir).mkdir()

    # Create model module dir
    model_dir = target_directory_path / config.model_module_dir
    template_path = TEMPLATES_DIR / 'custom'
    copy_tree_path(template_path / 'model', model_dir)

    examples_path = template_path / DEFAULT_EXAMPLES_FILENAME
    if examples_path.exists():
        copy_file_path(examples_path, target_directory_path / DEFAULT_EXAMPLES_FILENAME)

    # Write config
    with (target_directory_path / CONFIG_FILE).open('w') as config_file:
        yaml.dump(config.to_dict(), config_file)

    scaf = TrussHandle(target_directory_path)
    _update_truss_props(scaf, data_files, requirements_file)
    return scaf


def from_directory(truss_directory: str) -> TrussHandle:
    """Get a handle to a Truss. A Truss is a build context designed to be built
       as a container locally or uploaded into a baseten serving environment.

       Args:
           truss_directory (str): The local directory of an existing Truss
       Returns:
           TrussHandle
       """
    return TrussHandle(Path(truss_directory))


def cleanup():
    """
    Cleans up .truss directory.
    """
    build_folder_path = Path(
        Path.home(),
        '.truss'
    )
    if (build_folder_path.exists()):
        for obj in build_folder_path.glob('**/*'):
            if (not obj.name == 'config.yaml') and (obj.is_file()):
                os.remove(obj)
    return


def _update_truss_props(
    scaf: TrussHandle,
    data_files: List[str] = None,
    requirements_file: str = None,
):
    if data_files is not None:
        for data_file in data_files:
            scaf.add_data(data_file)

    if requirements_file is not None:
        scaf.update_requirements_from_file(requirements_file)


def kill_all():
    kill_containers({
        TRUSS: True
    })



from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, List

import yaml

from truss.types import ModelFrameworkType
from truss.validation import validate_secret_name

DEFAULT_MODEL_FRAMEWORK_TYPE = ModelFrameworkType.CUSTOM
DEFAULT_MODEL_TYPE = 'Model'
DEFAULT_MODEL_MODULE_DIR = 'model'
DEFAULT_MODEL_CLASS_FILENAME = 'model.py'
DEFAULT_MODEL_CLASS_NAME = 'Model'
DEFAULT_TRUSS_STRUCTURE_VERSION = '2.0'
DEFAULT_MODEL_INPUT_TYPE = 'Any'
DEFAULT_PYTHON_VERSION = 'py39'
DEFAULT_DATA_DIRECTORY = 'data'
DEFAULT_EXAMPLES_FILENAME = 'examples.yaml'

DEFAULT_CPU = '500m'
DEFAULT_MEMORY = '512m'
DEFAULT_USE_GPU = False


@dataclass
class Resources:
    cpu: str = DEFAULT_CPU
    memory: str = DEFAULT_MEMORY
    use_gpu: bool = DEFAULT_USE_GPU

    @staticmethod
    def from_dict(d):
        return Resources(
            cpu=d.get('cpu', DEFAULT_CPU),
            memory=d.get('memory', DEFAULT_MEMORY),
            use_gpu=d.get('use_gpu', DEFAULT_USE_GPU),
        )

    def to_dict(self):
        return {
            'cpu': self.cpu,
            'memory': self.memory,
            'use_gpu': self.use_gpu,
        }


@dataclass
class TrussConfig:
    model_framework: ModelFrameworkType = DEFAULT_MODEL_FRAMEWORK_TYPE
    model_type: str = DEFAULT_MODEL_TYPE
    # todo document model_name - ask phil
    model_name: str = None

    model_module_dir: str = DEFAULT_MODEL_MODULE_DIR
    model_class_filename: str = DEFAULT_MODEL_CLASS_FILENAME
    model_class_name: str = DEFAULT_MODEL_CLASS_NAME

    data_dir: str = DEFAULT_DATA_DIRECTORY

    # Python types for what the model expects as input
    input_type: str = DEFAULT_MODEL_INPUT_TYPE
    model_metadata: Dict[str, Any] = field(default_factory=dict)
    requirements: List[str] = field(default_factory=list)
    system_packages: List[str] = field(default_factory=list)
    environment_variables: Dict[str, str] = field(default_factory=dict)
    resources: Resources = field(default_factory=Resources)
    python_version: str = DEFAULT_PYTHON_VERSION
    examples_filename: str = DEFAULT_EXAMPLES_FILENAME
    secrets: Dict[str, str] = field(default_factory=dict)

    @staticmethod
    def from_dict(d):
        config = TrussConfig(
            model_type=d.get('model_type', DEFAULT_MODEL_TYPE),
            model_framework=ModelFrameworkType(d.get('model_framework', DEFAULT_MODEL_FRAMEWORK_TYPE.value)),
            model_module_dir=d.get('model_module_dir', DEFAULT_MODEL_MODULE_DIR),
            model_class_filename=d.get('model_class_filename', DEFAULT_MODEL_CLASS_FILENAME),
            model_class_name=d.get('model_class_name', DEFAULT_MODEL_CLASS_NAME),
            data_dir=d.get('data_dir', DEFAULT_DATA_DIRECTORY),
            input_type=d.get('input_type', DEFAULT_MODEL_INPUT_TYPE),
            model_metadata=d.get('model_metadata', {}),
            requirements=d.get('requirements', []),
            system_packages=d.get('system_packages', []),
            environment_variables=d.get('environment_variables', {}),
            resources=Resources.from_dict(d.get('resources', {})),
            python_version=d.get('python_version', DEFAULT_PYTHON_VERSION),
            model_name=d.get('model_name', None),
            examples_filename=d.get('examples_filename', DEFAULT_EXAMPLES_FILENAME),
            secrets=d.get('secrets', {}),
        )
        config.validate()
        return config

    @staticmethod
    def from_yaml(yaml_path: Path):
        with yaml_path.open() as yaml_file:
            return TrussConfig.from_dict(yaml.safe_load(yaml_file))

    def write_to_yaml_file(self, path: Path):
        with path.open('w') as config_file:
            yaml.dump(self.to_dict(), config_file)

    def to_dict(self):
        return {
            'model_type': self.model_type,
            'model_framework': self.model_framework.value,
            'model_module_dir': self.model_module_dir,
            'model_class_filename': self.model_class_filename,
            'model_class_name': self.model_class_name,
            'data_dir': self.data_dir,
            'input_type': self.input_type,
            'model_metadata': self.model_metadata,
            'requirements': self.requirements,
            'system_packages': self.system_packages,
            'environment_variables': self.environment_variables,
            'resources': self.resources.to_dict(),
            'python_version': self.python_version,
            'model_name': self.model_name,
            'examples_filename': self.examples_filename,
            'secrets': self.secrets,
        }

    def clone(self):
        return TrussConfig.from_dict(self.to_dict())

    def validate(self):
        for secret_name in self.secrets:
            validate_secret_name(secret_name)



import contextlib
import importlib
import os
import sys
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import tensorflow as tf
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier
from tensorflow.keras import layers
from tensorflow.keras.layers.experimental import preprocessing
from transformers import AutoModelWithLMHead, AutoTokenizer, pipeline

from truss.build import init

PYTORCH_MODEL_FILE_CONTENTS = """
import torch
import torch.nn as nn

class MyModel(nn.Module):
    def __init__(self):
        super(MyModel, self).__init__()
        self.main = torch.nn.Sequential(
            torch.nn.Linear(3, 1),
            torch.nn.Flatten(0, 1)
        )

    def forward(self, input):
        return self.main(input)

"""

PYTORCH_MODEL_FILE_WITH_NUMPY_IMPORT_CONTENTS = """
import numpy as np
import torch
import torch.nn as nn

class MyModel(nn.Module):
    def __init__(self):
        super(MyModel, self).__init__()
        self.main = torch.nn.Sequential(
            torch.nn.Linear(3, 1),
            torch.nn.Flatten(0, 1)
        )

    def forward(self, input):
        return self.main(input)

"""


PYTORCH_WITH_INIT_ARGS_MODEL_FILE_CONTENTS = """
import torch
import torch.nn as nn

class MyModel(nn.Module):
    def __init__(self, arg1, arg2, kwarg1=1, kwarg2=2):
        super(MyModel, self).__init__()
        self.arg1 = arg1
        self.arg2 = arg2
        self.kwarg1 = kwarg1
        self.kwarg2 = kwarg2

    def forward(self, input):
        return input

"""


CUSTOM_MODEL_CODE = """
class Model:
    def __init__(*args, **kwargs):
        pass

    def load(self):
        pass

    def predict(self, request):
        return [1 for i in request['inputs']]
"""

CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS = '''
class Model:
    def __init__(*args, **kwargs):
        pass

    def load(*args, **kwargs):
        pass

    def preprocess(self, request):
        # Adds 1 to all
        return {
            'inputs': [value + 1 for value in request['inputs']]
        }

    def predict(self, request):
        # Returns inputs as predictions
        return {
            'predictions': request['inputs'],
        }

    def postprocess(self, request):
        # Adds 2 to all
        return {
            'predictions': [value + 2 for value in request['predictions']],
        }
'''

CUSTOM_MODEL_CODE_FOR_GPU_TESTING = '''
import subprocess

class Model:
    def predict(self, request):
        process = subprocess.run(['nvcc','--version'], check=True, stdout=subprocess.PIPE, universal_newlines=True)
        cuda_version = process.stdout.split('\\n')[-2].split()[-1].split('/')[0].split('_')[-1]
        return {
            'predictions': [{'cuda_version': cuda_version}],
        }
'''

CUSTOM_MODEL_CODE_FOR_SECRETS_TESTING = '''
import subprocess

class Model:
    def __init__(self, secrets):
        self._secrets = secrets

    def predict(self, request):
        # Expects instance to be secret name and returns secret value as prediction
        secret_name = request['instances'][0]
        return {
            'predictions': [self._secrets[secret_name]],
        }
'''


# Doesn't implement load
NO_LOAD_CUSTOM_MODEL_CODE = """
class Model:
    def preprocess(self, request):
        return request

    def postprocess(self, request):
        return request

    def predict(self, request):
        return {
            'predictions': [1]
        }
"""


# Doesn't implement predict
NO_PREDICT_CUSTOM_MODEL_CODE = """
class MyModel:
    def load(self):
        pass
"""

# Doesn't implement preprocess
NO_PREPROCESS_CUSTOM_MODEL_CODE = """
class Model:
     def load(*args, **kwargs):
        pass

     def postprocess(self, request):
        # Adds 1 to all
        return {
            'predictions': [value + 1 for value in request['predictions']],
        }

     def predict(self, request):
        return {
            'predictions': request['inputs'],
        }
"""


# Doesn't implement postprocess
NO_POSTPROCESS_CUSTOM_MODEL_CODE = """
class Model:
     def load(*args, **kwargs):
        pass

     def preprocess(self, request):
        # Adds 1 to all
        return {
            'inputs': [value + 1 for value in request['inputs']],
        }

     def predict(self, request):
        return {
            'predictions': request['inputs'],
        }
"""

# Implements no params for init
NO_PARAMS_INIT_CUSTOM_MODEL_CODE = """
class Model:
     def __init__(self):
        pass

     def preprocess(self, request):
        return request

     def postporcess(self, request):
        return request

     def predict(self, request):
        return {
            'predictions': request['inputs'],
        }
"""


@pytest.fixture
def pytorch_model(tmp_path):
    return _pytorch_model_from_content(
        tmp_path,
        PYTORCH_MODEL_FILE_CONTENTS,
        model_module_name='my_model',
        model_class_name='MyModel',
        model_filename='my_model.py',
    )


@pytest.fixture
def pytorch_model_with_numpy_import(tmp_path):
    return _pytorch_model_from_content(
        tmp_path,
        PYTORCH_MODEL_FILE_WITH_NUMPY_IMPORT_CONTENTS,
        model_module_name='my_model',
        model_class_name='MyModel',
        model_filename='my_model.py',
    )


@pytest.fixture
def pytorch_model_init_args():
    return {'arg1': 1, 'arg2': 2, 'kwarg1': 3, 'kwarg2': 4}


@pytest.fixture
def pytorch_model_with_init_args(tmp_path, pytorch_model_init_args):
    f = tmp_path / 'my_model_with_init.py'
    f.write_text(PYTORCH_WITH_INIT_ARGS_MODEL_FILE_CONTENTS)

    sys.path.append(str(tmp_path))
    model_class = getattr(importlib.import_module('my_model_with_init'), 'MyModel')
    return model_class(**pytorch_model_init_args), f


@pytest.fixture
def custom_model_truss_dir(tmp_path):
    dir_path = tmp_path / 'custom_truss'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(CUSTOM_MODEL_CODE)
    return dir_path


@pytest.fixture
def no_preprocess_custom_model(tmp_path):
    dir_path = tmp_path / 'my_no_preprocess_model'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(NO_PREPROCESS_CUSTOM_MODEL_CODE)
    yield dir_path


@pytest.fixture
def no_postprocess_custom_model(tmp_path):
    dir_path = tmp_path / 'my_no_postprocess_model'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(NO_POSTPROCESS_CUSTOM_MODEL_CODE)
    yield dir_path


@pytest.fixture
def no_load_custom_model(tmp_path):
    dir_path = tmp_path / 'my_no_load_model'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(NO_LOAD_CUSTOM_MODEL_CODE)
    yield dir_path


@pytest.fixture
def no_params_init_custom_model(tmp_path):
    dir_path = tmp_path / 'my_no_params_init_load_model'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(NO_PARAMS_INIT_CUSTOM_MODEL_CODE)
    yield dir_path


@pytest.fixture
def useless_file(tmp_path):
    f = tmp_path / 'useless.py'
    f.write_text('')
    sys.path.append(str(tmp_path))
    return f


@contextlib.contextmanager
def temp_dir(directory):
    """A context to allow user to drop into the temporary
    directory created by the tmp_path fixture"""
    current_dir = os.getcwd()
    os.chdir(directory)
    try:
        yield
    finally:
        os.chdir(current_dir)


@pytest.fixture(scope="session")
def iris_dataset():
    return load_iris()


@pytest.fixture(scope="session")
def sklearn_rfc_model(iris_dataset):
    data_x = iris_dataset['data']
    data_y = iris_dataset['target']
    data_x = pd.DataFrame(data_x)
    rfc_model = RandomForestClassifier()
    rfc_model.fit(data_x, data_y)
    return rfc_model


@pytest.fixture(scope="session")
def mpg_dataset():
    # url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/auto-mpg/auto-mpg.data'
    auto_mpg_data = Path(__file__).parent.parent / 'test_data' / 'auto-mpg.data'
    column_names = ['MPG', 'Cylinders', 'Displacement', 'Horsepower', 'Weight',
                    'Acceleration', 'Model Year', 'Origin']

    raw_dataset = pd.read_csv(
        str(auto_mpg_data),
        names=column_names,
        na_values='?',
        comment='\t',
        sep=' ',
        skipinitialspace=True
    )

    dataset = raw_dataset.copy()
    dataset.isna().sum()
    dataset = dataset.dropna()
    dataset['Origin'] = dataset['Origin'].map({1: 'USA', 2: 'Europe', 3: 'Japan'})
    dataset = pd.get_dummies(dataset, columns=['Origin'], prefix='', prefix_sep='')
    return dataset


@pytest.fixture(scope="session")
def keras_mpg_model(mpg_dataset):
    train_dataset = mpg_dataset.sample(frac=0.8, random_state=0)
    train_features = train_dataset.copy()
    train_labels = train_features.pop('MPG')
    normalizer = preprocessing.Normalization(axis=-1)
    normalizer.adapt(np.array(train_features))
    linear_model = tf.keras.Sequential([
        normalizer,
        layers.Dense(units=1)
    ])

    linear_model.compile(
        optimizer=tf.optimizers.Adam(learning_rate=0.1),
        loss='mean_absolute_error'
    )

    linear_model.fit(
        train_features, train_labels,
        epochs=5,
        # suppress logging
        verbose=0,
        # Calculate validation results on 20% of the training data
        validation_split=0.2
    )
    return linear_model


@pytest.fixture(scope="session")
def huggingface_transformer_t5_small_model():
    tokenizer = AutoTokenizer.from_pretrained('t5-small')
    model = AutoModelWithLMHead.from_pretrained('t5-small')
    return pipeline('text2text-generation', model=model, tokenizer=tokenizer)


@pytest.fixture
def custom_model_truss_dir_with_pre_and_post_no_example(tmp_path):
    dir_path = tmp_path / 'custom_truss_with_pre_post_no_example'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS)
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_with_pre_and_post(tmp_path):
    dir_path = tmp_path / 'custom_truss_with_pre_post'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS)
    sc.update_examples({
        'example1': {
            'inputs': [[0]]
        }
    })
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_with_pre_and_post_str_example(tmp_path):
    dir_path = tmp_path / 'custom_truss_with_pre_post_str_example'
    sc = init(str(dir_path))
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(CUSTOM_MODEL_CODE_WITH_PRE_AND_POST_PROCESS)
    sc.update_examples({
        'example1': {
            'inputs': [{
                'image_url' : 'https://github.com/pytorch/hub/raw/master/images/dog.jpg'
            }]
        }
    })
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_for_gpu(tmp_path):
    dir_path = tmp_path / 'custom_truss'
    sc = init(str(dir_path))
    sc.enable_gpu()
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(CUSTOM_MODEL_CODE_FOR_GPU_TESTING)
    yield dir_path


@pytest.fixture
def custom_model_truss_dir_for_secrets(tmp_path):
    dir_path = tmp_path / 'custom_truss'
    sc = init(str(dir_path))
    sc.add_secret('secret_name', 'default_secret_value')
    with sc.spec.model_class_filepath.open('w') as file:
        file.write(CUSTOM_MODEL_CODE_FOR_SECRETS_TESTING)
    yield dir_path


def _pytorch_model_from_content(
    path: Path,
    content: str,
    model_module_name: str,
    model_class_name: str,
    model_filename: str,
):
    f = path / model_filename
    f.write_text(content)

    sys.path.append(str(path))
    model_class = getattr(importlib.import_module(model_module_name), model_class_name)
    return model_class(), f



from typing import Dict, List


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs['data_dir']
        self._config = kwargs['config']
        self._model = None

    def load(self):
        # Load model here and assign to self._model.
        pass

    def preprocess(self, request: Dict) -> Dict:
        """
        Incorporate pre-processing required by the model if desired here.

        These might be feature transformations that are tightly coupled to the model.
        """
        return request

    def postprocess(self, request: Dict) -> Dict:
        """
        Incorporate post-processing required by the model if desired here.
        """
        return request

    def predict(self, request: Dict) -> Dict[str, List]:
        response = {}
        inputs = request['inputs'] # noqa
        # Invoke model and calculate predictions here.
        response['predictions'] = []
        return response



import pytest

from truss.validation import validate_secret_name


@pytest.mark.parametrize('secret_name, should_error', [
    (None, True),
    (1, True),
    ('', True),
    ('.', True),
    ('..', True),
    ('a' * 253, False),
    ('a' * 254, True),
    ('-', False),
    ('-.', False),
    ('a-.', False),
    ('-.a', False),
    ('a-foo', False),
    ('a.foo', False),
    ('.foo', False),
    ('x\\', True),
    ('a_b', False),
    ('_a', False),
    ('a_', False),
])
def test_validate_secret_name(secret_name, should_error):
    does_error = False
    try:
        validate_secret_name(secret_name)
    except:  # noqa
        does_error = True

    assert does_error == should_error



import tempfile
from pathlib import Path

import pytest
import requests
from python_on_whales import docker
from tenacity import Retrying, stop_after_attempt, wait_fixed

from truss.constants import CONFIG_FILE
from truss.contexts.image_builder.image_builder import ImageBuilderContext
from truss.contexts.local_loader.load_local import LoadLocal
from truss.model_frameworks.keras import Keras
from truss.truss_config import TrussConfig


def test_to_truss(keras_mpg_model):
    with tempfile.TemporaryDirectory(dir='.') as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, 'truss')
        framework = Keras()
        framework.to_truss(keras_mpg_model, truss_dir)

        # Assertions
        config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
        assert config.model_class_filename == 'model.py'
        assert config.model_class_name == 'Model'
        assert config.model_framework == framework.typ()
        assert config.model_type == 'Model'
        assert config.python_version.startswith('py3')
        assert len(config.requirements) != 0

        model_metadata = config.model_metadata
        assert model_metadata['model_binary_dir'] == 'model'

        assert (truss_dir / 'data' / 'model').exists()
        assert (truss_dir / 'model' / 'model.py').exists()


def test_run_truss(keras_mpg_model):
    with tempfile.TemporaryDirectory(dir='.') as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, 'truss')
        sklearn_framework = Keras()
        sklearn_framework.to_truss(keras_mpg_model, truss_dir)
        model = LoadLocal.run(truss_dir)
        result = model.predict({'inputs': [[0, 0, 0, 0, 0, 0, 0, 0, 0]]})
        predictions = result['predictions']
        assert len(predictions) == 1
        assert len(predictions[0]) == 1


@pytest.mark.integration
def test_run_image(keras_mpg_model):
    with tempfile.TemporaryDirectory(dir='.') as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, 'truss')
        framework = Keras()
        framework.to_truss(keras_mpg_model, truss_dir)
        tag = f'test-{framework.typ().value}-model:latest'
        image = ImageBuilderContext.run(truss_dir).build_image(tag=tag)
        assert image.repo_tags == [tag]
        container = docker.run(tag, publish=[[8080, 8080]], detach=True)
        try:
            for attempt in Retrying(stop=stop_after_attempt(10), wait=wait_fixed(2)):
                with attempt:
                    resp = requests.post('http://localhost:8080/v1/models/model:predict', json={
                        'inputs': [[0, 0, 0, 0, 0, 0, 0, 0, 0]],
                    })
        finally:
            docker.kill(container)
        predictions = resp.json()['predictions']
        assert len(predictions) == 1
        assert len(predictions[0]) == 1




class Error(Exception):
    """Base Error"""
    def __init__(self, message):
        super(Error, self).__init__(message)
        self.message = message


class InferenceError(Error):
    """Error raised when model inference fails.
    """



from typing import Dict, List

import numpy as np
from tensorflow import keras


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs['data_dir']
        config = kwargs['config']
        model_metadata = config['model_metadata']
        self._model_binary_dir = model_metadata['model_binary_dir']
        self._model = None

    def load(self):
        self._model = keras.models.load_model(str(self._data_dir / self._model_binary_dir))

    def preprocess(self, request: Dict) -> Dict:
        """
        Incorporate pre-processing required by the model if desired here.

        These might be feature transformations that are tightly coupled to the model.
        """
        return request

    def postprocess(self, request: Dict) -> Dict:
        """
        Incorporate post-processing required by the model if desired here.
        """
        return request

    def predict(self, request: Dict) -> Dict[str, List]:
        response = {}
        inputs = request['inputs']
        inputs = np.array(inputs)
        result = self._model.predict(inputs).tolist()
        response['predictions'] = result
        return response



import datetime
import decimal
import json
import uuid

import msgpack
import msgpack_numpy as mp_np
import numpy as np


# mostly cribbed from django.core.serializer.DjangoJSONEncoder
def truss_msgpack_encoder(obj, chain=None):
    if isinstance(obj, datetime.datetime):
        r = obj.isoformat()
        if r.endswith('+00:00'):
            r = r[:-6] + 'Z'
        return {b'__dt_datetime_iso__': True, b'data': r}
    elif isinstance(obj, datetime.date):
        r = obj.isoformat()
        return {b'__dt_date_iso__': True, b'data': r}
    elif isinstance(obj, datetime.time):
        if obj.utcoffset() is not None:
            raise ValueError("Cannot represent timezone-aware times.")
        r = obj.isoformat()
        return {b'__dt_time_iso__': True, b'data': r}
    elif isinstance(obj, datetime.timedelta):
        return {b'__dt_timedelta__': True, b'data': (obj.days, obj.seconds, obj.microseconds)}
    elif isinstance(obj, decimal.Decimal):
        return {b'__decimal__': True, b'data': str(obj)}
    elif isinstance(obj, uuid.UUID):
        return {b'__uuid__': True, b'data': str(obj)}
    else:
        return obj if chain is None else chain(obj)


def truss_msgpack_decoder(obj, chain=None):
    try:
        if b'__dt_datetime_iso__' in obj:
            return datetime.datetime.fromisoformat(obj[b'data'])
        elif b'__dt_date_iso__' in obj:
            return datetime.date.fromisoformat(obj[b'data'])
        elif b'__dt_time_iso__' in obj:
            return datetime.time.fromisoformat(obj[b'data'])
        elif b'__dt_timedelta__' in obj:
            days, seconds, microseconds = obj[b'data']
            return datetime.timedelta(days=days, seconds=seconds, microseconds=microseconds)
        elif b'__decimal__' in obj:
            return decimal.Decimal(obj[b'data'])
        elif b'__uuid__' in obj:
            return uuid.UUID(obj[b'data'])
        else:
            return obj if chain is None else chain(obj)
    except KeyError:
        return obj if chain is None else chain(obj)


# this json object is JSONType + np.array + datetime
def is_truss_serializable(obj):
    # basic JSON types
    if isinstance(obj, (str, int, float, bool, type(None), dict, list)):
        return True
    elif isinstance(obj, (datetime.datetime, datetime.date, datetime.time, datetime.timedelta)):
        return True
    elif isinstance(obj, np.ndarray):
        return True
    else:
        return False


def truss_msgpack_serialize(obj):
    return msgpack.packb(obj, default=lambda x: truss_msgpack_encoder(x, chain=mp_np.encode))


def truss_msgpack_deserialize(obj):
    return msgpack.unpackb(obj, object_hook=lambda x: truss_msgpack_decoder(x, chain=mp_np.decode))


class DeepNumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(DeepNumpyEncoder, self).default(obj)



import copy
import glob
import logging
from dataclasses import replace
from pathlib import Path
from typing import Callable, Dict, List, Union

import numpy as np
import requests
import yaml
from tenacity import Retrying, stop_after_attempt, wait_fixed

from truss.constants import TRUSS, TRUSS_DIR, TRUSS_MODIFIED_TIME
from truss.contexts.image_builder.image_builder import ImageBuilderContext
from truss.contexts.local_loader.load_local import LoadLocal
from truss.docker import (Docker, get_container_logs, get_containers,
                          get_images, get_urls_from_container, kill_containers)
from truss.local.local_config_handler import LocalConfigHandler
from truss.readme_generator import generate_readme
from truss.truss_config import TrussConfig
from truss.truss_spec import TrussSpec
from truss.utils import (copy_file_path, copy_tree_path,
                         get_max_modified_time_of_dir)
from truss.validation import validate_secret_name

logger = logging.getLogger(__name__)


class TrussHandle:
    def __init__(self, truss_dir: Path) -> None:
        self._truss_dir = truss_dir
        self._spec = TrussSpec(truss_dir)

    @property
    def spec(self) -> TrussSpec:
        return self._spec

    def server_predict(self, request: dict):
        """Run the prediction flow locally."""
        model = LoadLocal.run(self._truss_dir)
        return _prediction_flow(model, request)

    def build_docker_image(self, build_dir: Path = None, tag: str = None):
        """Builds docker image"""
        images = self.get_docker_images_from_label()
        if images and isinstance(images, list):
            return images[0]
        build_dir_path = Path(build_dir) if build_dir is not None else None
        image_builder = ImageBuilderContext.run(self._truss_dir)
        return image_builder.build_image(build_dir_path, tag, labels=self._get_labels())

    def docker_run(
        self,
        build_dir: Path = None,
        tag: str = None,
        local_port: int = 8080,
        detach=True,
    ):
        """
        Builds a docker image and runs it as a container.

        Returns:
            Container, which can be used to get information about the running,
            including its id. The id can be used to kill the container.
        """
        image = self.build_docker_image(build_dir=build_dir, tag=tag)
        built_tag = image.repo_tags[0]
        labels = self._get_labels()
        labels.update({TRUSS : True})
        secrets_mount_dir_path = _prepare_secrets_mount_dir()
        container = Docker.client().run(
            built_tag,
            publish=[[local_port, 8080]],
            detach=detach,
            labels=labels,
            mounts=[[
                'type=bind',
                f'src={str(secrets_mount_dir_path)}',
                'target=/secrets',
            ]],
            gpus='all' if self._spec.config.resources.use_gpu else None,
        )
        model_base_url = f'http://localhost:{local_port}/'
        try:
            _wait_for_model_server(model_base_url)
        except Exception as e:
            for log in self.container_logs():
                logging.info(log)
            raise e
        logger.info(f'Model server started on port {local_port}, docker container id {container.id}')
        return container

    def docker_predict(
        self,
        request: dict,
        build_dir: Path = None,
        tag: str = None,
        local_port: int = 8080,
        detach: bool = True,
    ):
        """
        Builds docker image, runs that as a docker container
        and makes a prediction request to the server running on the container.
        Kills the container afterwards. Mostly useful for testing.
        """
        containers = self.get_docker_containers_from_labels()
        if containers:
            container = containers[0]
        else:
            container = self.docker_run(build_dir, tag, local_port=local_port, detach=detach)
        model_base_url = get_urls_from_container(container)[0]
        resp = requests.post(f'{model_base_url}/v1/models/model:predict', json=request)
        resp.raise_for_status()
        return resp.json()

    def docker_build_setup(self, build_dir: Path = None):
        """
        Set up a directory to build docker image from.

        Returns:
            docker build command.
        """
        image_builder = ImageBuilderContext.run(self._truss_dir)
        image_builder.prepare_image_build_dir(build_dir)
        return image_builder.docker_build_command(build_dir)

    def add_python_requirement(self, python_requirement: str):
        """Add a python requirement to truss model's config."""
        self._update_config(lambda conf: replace(
            conf,
            requirements=[*conf.requirements, python_requirement]))

    def add_environment_variable(self, env_var_name: str, env_var_value: str):
        """Add an environment variable to truss model's config."""
        if not env_var_value:
            logger.info("Enviroment value should not empty or none!")
            return

        self._update_config(lambda conf: replace(
            conf,
            environment_variables={
                **conf.environment_variables,
                env_var_name: env_var_value,
            },
        ))

    def add_secret(self, secret_name: str, default_secret_value: str = ''):
        validate_secret_name(secret_name)
        self._update_config(lambda conf: replace(
            conf,
            secrets={
                **conf.secrets,
                secret_name: default_secret_value,
            },
        ))

    def update_requirements(self, requirements: List[str]):
        """Update requirements in truss model's config.

        Replaces requirements in truss model's config with the provided list.
        """
        self._update_config(lambda conf: replace(conf, requirements=requirements))

    def update_requirements_from_file(self, requirements_filepath: str):
        """Update requirements in truss model's config.

        Replaces requirements in truss model's config with those from the file
        at the given path.
        """
        with Path(requirements_filepath).open() as req_file:
            self.update_requirements([line.strip() for line in req_file.readlines()])

    def add_system_package(self, system_package: str):
        """Add a system package requirement to truss model's config."""
        self._update_config(lambda conf: replace(
            conf,
            system_packages=[*conf.system_packages, system_package]))

    def add_data(self, file_dir_or_glob: str):
        """Add data to a truss model.

        Accepts a file path, a directory path or a glob. Everything is copied
        under the truss model's data directory.
        """
        item = file_dir_or_glob
        item_path = Path(item)
        if item_path.is_dir():
            copy_tree_path(item_path, self._spec.data_dir / item_path.name)
        else:
            filenames = glob.glob(item)
            for filename in filenames:
                filepath = Path(filename)
                copy_file_path(filepath, self._spec.data_dir / filepath.name)

    def examples(self) -> Dict[str, Dict]:
        """List truss model's examples.

        Examples are a simple `name to input` dictionary.
        """
        return self._spec.examples

    def update_examples(self, examples: Dict[str, Dict]):
        """Update truss model's examples.

        Existing examples are replaced whole with the given ones.
        """
        with self._spec.examples_path.open('w') as examples_file:
            examples_file.write(yaml.dump(examples))

    def example(self, name_or_index: Union[str, int]) -> Dict:
        """Return lookup an example by name or index.

        Index is 0 based. e.g. example(0) returns the first example.
        """
        if isinstance(name_or_index, str):
            return self.examples()[name_or_index]
        return list(self.examples().values())[name_or_index]

    def add_example(self, example_name: str, example: dict):
        """Add example for truss model.

        If the example with the given name already exists then it is overwritten.
        """
        examples = copy.deepcopy(self.examples())
        examples[example_name] = example
        self.update_examples(examples)

    def get_docker_images_from_label(self):
        return get_images(self._get_labels())

    def get_docker_containers_from_labels(self, all=False):
        return sorted(get_containers(self._get_labels(), all=all), key=lambda c: c.created)

    def kill_container(self):
        kill_containers(self._get_labels())

    def container_logs(self):
        containers = self.get_docker_containers_from_labels(all=True)
        if not containers:
            raise ValueError("No Container is running for truss!")
        return get_container_logs(containers[-1])

    def enable_gpu(self):
        """Enable gpu use for given model.

        This is suggestive, model serving environment may still use cpu, e.g. if
        the setup doesn't have access to a GPU.

        Note that truss would typically use a larger docker base image when this
        is enabled, for example to include the cuda libraries.
        """
        def enable_gpu_fn(conf: TrussConfig):
            new_resources = replace(conf.resources, use_gpu=True)
            return replace(conf, resources=new_resources)

        self._update_config(enable_gpu_fn)

    def _get_labels(self):
        return {
            TRUSS_MODIFIED_TIME : get_max_modified_time_of_dir(
                self._truss_dir
            ),
            TRUSS_DIR: self._truss_dir,
        }

    def _update_config(self, update_config_fn: Callable[[TrussConfig], TrussConfig]):
        config = update_config_fn(self._spec.config)
        config.write_to_yaml_file(self._spec.config_path)
        # reload spec
        self._spec = TrussSpec(self._truss_dir)

    def get_urls_from_truss(self):
        urls = []
        containers = self.get_docker_containers_from_labels()
        for container in containers:
            urls.extend(get_urls_from_container(container))
        return urls

    def generate_readme(self):
        return generate_readme(self._spec)


def _prediction_flow(model, request: dict):
    """This flow attempts to mimic the request life-cycle of a kfserving server"""
    _validate_request_input(request)
    _map_instances_inputs(request)
    if hasattr(model, 'preprocess'):
        request = model.preprocess(request)
    response = model.predict(request)
    if hasattr(model, 'postprocess'):
        response = model.postprocess(response)
    return response


def _map_instances_inputs(request: dict):
    # TODO(pankaj) Share this code with baseten deployed code
    if "instances" in request and "inputs" not in request:
        request["inputs"] = request["instances"]
    elif "inputs" in request and "instances" not in request:
        request["instances"] = request["inputs"]
    return request


def _validate_request_input(request: dict):
    # TODO(pankaj) Should these checks be there?
    if _is_invalid_list_input_prop(request, 'instances') \
            or _is_invalid_list_input_prop(request, 'inputs'):
        raise Exception(
            reason="Expected \"instances\" or \"inputs\" to be a list"
        )


def _is_invalid_list_input_prop(request: dict, prop: str):
    return prop in request and not _is_valid_list_type(request[prop])


def _is_valid_list_type(obj) -> bool:
    return isinstance(obj, (list, np.ndarray))


def _wait_for_model_server(url: str):
    for attempt in Retrying(stop=stop_after_attempt(10), wait=wait_fixed(2)):
        with attempt:
            resp = requests.get(url)
            resp.raise_for_status()


def _prepare_secrets_mount_dir() -> Path:
    LocalConfigHandler.sync_secrets_mount_dir()
    return LocalConfigHandler.secrets_dir_path()




import copy
import inspect
from pathlib import Path
from typing import Dict

from truss.contexts.local_loader.model_module_loader import \
    model_class_module_loaded
from truss.contexts.truss_context import TrussContext
from truss.local.local_config_handler import LocalConfigHandler
from truss.truss_spec import TrussSpec


class LoadLocal(TrussContext):
    """Loads a Truss model locally.

    The loaded model can be used to make predictions for quick testing.
    Runs in the current pip environment directly. Assumes all requirements and
    system packages are already installed.
    """

    @staticmethod
    def run(truss_dir: Path):
        spec = TrussSpec(truss_dir)
        with model_class_module_loaded(str(truss_dir), spec.model_module_fullname) as module:
            model_class = getattr(module, spec.model_class_name)
            model_class_signature = inspect.signature(model_class)
            model_init_params = {}
            if _signature_accepts_keyword_arg(model_class_signature, 'config'):
                model_init_params['config'] = spec.config.to_dict()
            if _signature_accepts_keyword_arg(model_class_signature, 'data_dir'):
                model_init_params['data_dir'] = truss_dir / 'data'
            if _signature_accepts_keyword_arg(model_class_signature, 'secrets'):
                model_init_params['secrets'] = _prepare_secrets(spec)
            model = model_class(**model_init_params)
            if hasattr(model, 'load'):
                model.load()
            return model


def _signature_accepts_keyword_arg(signature: inspect.Signature, kwarg: str) -> bool:
    return kwarg in signature.parameters or _signature_accepts_kwargs(signature)


def _signature_accepts_kwargs(signature: inspect.Signature) -> bool:
    for param in signature.parameters.values():
        if param.kind == inspect.Parameter.VAR_KEYWORD:
            return True
    return False


def _prepare_secrets(spec: TrussSpec) -> Dict[str, str]:
    secrets = copy.deepcopy(spec.secrets)
    local_secerts = LocalConfigHandler.get_config().secrets
    for secret_name in secrets:
        if secret_name in local_secerts:
            secrets[secret_name] = local_secerts[secret_name]
    return secrets



import tempfile
from pathlib import Path

import numpy as np
import pytest

from truss.constants import CONFIG_FILE
from truss.contexts.local_loader.load_local import LoadLocal
from truss.model_frameworks.sklearn import SKLearn
from truss.truss_config import TrussConfig
from truss.truss_handle import TrussHandle
from truss.types import ModelFrameworkType


def test_to_truss(sklearn_rfc_model):
    with tempfile.TemporaryDirectory(dir='.') as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, 'truss')
        sklearn_framework = SKLearn()
        sklearn_framework.to_truss(sklearn_rfc_model, truss_dir)

        # Assertions
        config = TrussConfig.from_yaml(truss_dir / CONFIG_FILE)
        assert config.model_class_filename == 'model.py'
        assert config.model_class_name == 'Model'
        assert config.model_framework == ModelFrameworkType.SKLEARN
        assert config.model_type == 'Model'
        assert config.python_version.startswith('py3')
        assert len(config.requirements) != 0

        model_metadata = config.model_metadata
        assert model_metadata['model_binary_dir'] == 'model'
        assert model_metadata['supports_predict_proba']

        assert (truss_dir / 'data' / 'model' / 'model.joblib').exists()
        assert (truss_dir / 'model' / 'model.py').exists()


def test_run_truss(sklearn_rfc_model):
    with tempfile.TemporaryDirectory(dir='.') as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, 'truss')
        sklearn_framework = SKLearn()
        sklearn_framework.to_truss(sklearn_rfc_model, truss_dir)
        model = LoadLocal.run(truss_dir)
        predictions = model.predict({'inputs': [[0, 0, 0, 0]]})
        assert len(predictions['probabilities']) == 1
        assert len(predictions['probabilities'][0]) == 3


@pytest.mark.integration
def test_run_image(sklearn_rfc_model):
    with tempfile.TemporaryDirectory(dir='.') as tmp_work_dir:
        truss_dir = Path(tmp_work_dir, 'truss')
        sklearn_framework = SKLearn()
        sklearn_framework.to_truss(sklearn_rfc_model, truss_dir)
        tr = TrussHandle(truss_dir)
        predictions = tr.docker_predict(
            {'inputs': [[0, 0, 0, 0]]},
            local_port=8090,
        )
        assert len(predictions['probabilities']) == 1
        assert np.shape(predictions['probabilities']) == (1, 3)



import yaml
from common.truss_server import TrussServer
from model_wrapper import ModelWrapper

CONFIG_FILE = 'config.yaml'


if __name__ == '__main__':
    with open(CONFIG_FILE, encoding='utf-8') as config_file:
        config = yaml.safe_load(config_file)
        model = ModelWrapper(config)
        model.load()
        TrussServer(workers=1).start([model])



import pytest

from truss.constants import PYTORCH
from truss.model_inference import (PYTORCH_REQ_MODULE_NAME,
                                   TENSORFLOW_REQ_MODULE_NAME,
                                   _get_entries_for_packages,
                                   infer_model_information,
                                   validate_provided_parameters_with_model)

SAMPLE_PIP_FREEZE_OUTPUT = [
    'tensorflow==2.9.1',
]


def test_pytorch_init_arg_validation(pytorch_model_with_init_args, pytorch_model_init_args):
    pytorch_model_with_init_args, _ = pytorch_model_with_init_args
    # Validates with args and kwargs
    validate_provided_parameters_with_model(pytorch_model_with_init_args.__class__, pytorch_model_init_args)

    # Errors if bad args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(pytorch_model_with_init_args.__class__, {'foo': 'bar'})

    # Validates with only args
    copied_args = pytorch_model_init_args.copy()
    copied_args.pop('kwarg1')
    copied_args.pop('kwarg2')
    validate_provided_parameters_with_model(pytorch_model_with_init_args, copied_args)

    # Requires all args
    with pytest.raises(ValueError):
        validate_provided_parameters_with_model(pytorch_model_with_init_args, {})


def test_infer_model_information(pytorch_model_with_init_args):
    model_info = infer_model_information(pytorch_model_with_init_args[0])
    assert model_info.model_framework == PYTORCH
    assert model_info.model_type == 'MyModel'


@pytest.mark.parametrize(
    'pip_freeze_output, desired_reqs, expected_req',
    [
        (['tensorflow==2.9.1+abc'], TENSORFLOW_REQ_MODULE_NAME, {'tensorflow': 'tensorflow==2.9.1'}),
        (['tensorflow==2.9.1'], TENSORFLOW_REQ_MODULE_NAME, {'tensorflow': 'tensorflow==2.9.1'}),
        (['tensorflow==2.9.1', 'dummy==a.b.c'], TENSORFLOW_REQ_MODULE_NAME, {'tensorflow': 'tensorflow==2.9.1'}),
        (['dummy==a.b.c'], TENSORFLOW_REQ_MODULE_NAME, {}),
        (['torch==1.12.0'], PYTORCH_REQ_MODULE_NAME, {'torch': 'torch==1.12.0'}),
    ],
)
def test_get_entries_for_packages(pip_freeze_output, desired_reqs, expected_req):
    entries = _get_entries_for_packages(pip_freeze_output, desired_reqs)
    assert entries == expected_req



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['truss',
 'truss.contexts',
 'truss.contexts.image_builder',
 'truss.contexts.local_loader',
 'truss.local',
 'truss.model_frameworks',
 'truss.templates',
 'truss.templates.custom.model',
 'truss.templates.huggingface_transformer.model',
 'truss.templates.keras.model',
 'truss.templates.pytorch.model',
 'truss.templates.server',
 'truss.templates.server.common',
 'truss.templates.sklearn.model',
 'truss.tests',
 'truss.tests.contexts.local_loader',
 'truss.tests.local',
 'truss.tests.model_frameworks',
 'truss.tests.templates.core.server',
 'truss.tests.templates.core.server.common']

package_data = \
{'': ['*'],
 'truss': ['test_data/*'],
 'truss.templates': ['custom/*',
                     'docs/*',
                     'huggingface_transformer/*',
                     'keras/*',
                     'pytorch/*',
                     'sklearn/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'PyYAML>=6.0,<7.0',
 'msgpack-numpy>=0.4.7.1',
 'msgpack>=1.0.2',
 'numpy>=1.18',
 'packaging>=20.9,<21.0',
 'python-json-logger>=2.0.2',
 'python-on-whales>=0.46.0,<0.47.0',
 'tenacity>=8.0.1,<9.0.0']

entry_points = \
{'console_scripts': ['truss = truss.cli:cli_group']}

setup_kwargs = {
    'name': 'truss',
    'version': '0.0.25',
    'description': '',
    'long_description': None,
    'author': 'Pankaj Gupta',
    'author_email': 'pankaj@baseten.co',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



from typing import Dict, List

MODEL_BASENAME = 'model'
MODEL_EXTENSIONS = ['.joblib', '.pkl', '.pickle']


class Model:
    def __init__(self, **kwargs) -> None:
        self._data_dir = kwargs['data_dir']
        config = kwargs['config']
        model_metadata = config['model_metadata']
        self._supports_predict_proba = model_metadata['supports_predict_proba']
        self._model_binary_dir = model_metadata['model_binary_dir']
        self._model = None

    def load(self):
        import joblib
        model_binary_dir_path = self._data_dir / self._model_binary_dir
        paths = [(model_binary_dir_path / MODEL_BASENAME).with_suffix(model_extension)
                 for model_extension in MODEL_EXTENSIONS]
        model_file_path = next(path for path in paths if path.exists())
        self._model = joblib.load(model_file_path)

    def preprocess(self, request: Dict) -> Dict:
        """
        Incorporate pre-processing required by the model if desired here.

        These might be feature transformations that are tightly coupled to the model.
        """
        return request

    def postprocess(self, request: Dict) -> Dict:
        """
        Incorporate post-processing required by the model if desired here.
        """
        return request

    def predict(self, request: Dict) -> Dict[str, List]:
        response = {}
        inputs = request['inputs']
        result = self._model.predict(inputs)
        response['predictions'] = result
        if self._supports_predict_proba:
            response['probabilities'] = self._model.predict_proba(inputs).tolist()
        return response



import tempfile
from contextlib import contextmanager
from pathlib import Path

import numpy as np

from truss.build import cleanup, init, mk_truss
from truss.truss_spec import TrussSpec


def test_scaffold_init(tmp_path):
    dir_name = str(tmp_path)
    init(dir_name)
    spec = TrussSpec(Path(dir_name))
    assert spec.model_module_dir.exists()
    assert spec.data_dir.exists()
    assert spec.truss_dir == tmp_path
    assert spec.config_path.exists()


def test_scaffold_init_with_data_file_and_requirements_file(tmp_path):
    dir_path = tmp_path / 'scaffold'
    dir_name = str(dir_path)
    data_file_path = tmp_path / 'data.txt'
    with data_file_path.open('w') as data_file:
        data_file.write('test')
    req_file_path = tmp_path / 'requirements.txt'
    requirements = [
        'tensorflow==2.3.1',
        'uvicorn==0.12.2',
    ]
    with req_file_path.open('w') as req_file:
        for req in requirements:
            req_file.write(f'{req}\n')

    init(dir_name, data_files=[str(data_file_path)], requirements_file=str(req_file_path))
    spec = TrussSpec(Path(dir_name))
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert (spec.data_dir / 'data.txt').exists()
    assert spec.requirements == requirements


def test_scaffold(sklearn_rfc_model, tmp_path):
    dir_path = tmp_path / 'scaffold'
    data_file_path = tmp_path / 'data.txt'
    with data_file_path.open('w') as data_file:
        data_file.write('test')
    req_file_path = tmp_path / 'requirements.txt'
    requirements = [
        'tensorflow==2.3.1',
        'uvicorn==0.12.2',
    ]
    with req_file_path.open('w') as req_file:
        for req in requirements:
            req_file.write(f'{req}\n')
    scaf = mk_truss(
        sklearn_rfc_model,
        target_directory=dir_path,
        data_files=[str(data_file_path)],
        requirements_file=str(req_file_path),
    )
    spec = scaf.spec
    assert spec.model_module_dir.exists()
    assert spec.truss_dir == dir_path
    assert spec.config_path.exists()
    assert spec.data_dir.exists()
    assert (spec.data_dir / 'data.txt').exists()
    assert spec.requirements == requirements


def test_scaffold_sklearn_predict(sklearn_rfc_model):
    with _model_server_predict(sklearn_rfc_model, {'inputs': [[0, 0, 0, 0]]}) as result:
        assert 'predictions' in result
        assert 'probabilities' in result
        probabilities = result['probabilities']
        assert np.shape(probabilities) == (1, 3)


def test_scaffold_keras_predict(keras_mpg_model):
    with _model_server_predict(
        keras_mpg_model,
        {'inputs': [[0, 0, 0, 0, 0, 0, 0, 0, 0]]},
    ) as result:
        assert 'predictions' in result
        predictions = result['predictions']
        assert np.shape(predictions) == (1, 1)


def test_scaffold_pytorch_predict(pytorch_model):
    model = pytorch_model[0]
    with _model_server_predict(
        model,
        {'inputs': [[0, 0, 0]]},
    ) as result:
        assert 'predictions' in result
        assert len(result['predictions']) == 1


def test_scaffold_huggingface_transformer_predict(huggingface_transformer_t5_small_model):
    with _model_server_predict(
        huggingface_transformer_t5_small_model,
        {'inputs': ['My name is Sarah and I live in London']},
    ) as result:
        print(result)
        assert 'predictions' in result
        predictions = result['predictions']
        assert len(predictions) == 1
        assert predictions[0]['generated_text'].startswith('Mein Name')


def test_cleanup(sklearn_rfc_model, tmp_path):
    data_file_path = tmp_path / 'data.txt'
    with data_file_path.open('w') as data_file:
        data_file.write('test')
    req_file_path = tmp_path / 'requirements.txt'
    requirements = [
        'tensorflow==2.3.1',
        'uvicorn==0.12.2',
    ]
    with req_file_path.open('w') as req_file:
        for req in requirements:
            req_file.write(f'{req}\n')
    _ = mk_truss(
        sklearn_rfc_model,
        data_files=[str(data_file_path)],
        requirements_file=str(req_file_path),
    )
    cleanup()
    build_folder_path = Path(
        Path.home(),
        '.truss'
    )
    directory = list(build_folder_path.glob("**/*"))
    files = [obj.name for obj in directory if obj.is_file()]
    unique_files = set(files)
    assert build_folder_path.exists()
    assert unique_files == {'config.yaml'}


@contextmanager
def _model_server_predict(model, model_input):
    with tempfile.TemporaryDirectory() as dir_name:
        sc = mk_truss(model, target_directory=dir_name)
        result = sc.server_predict(model_input)
        yield result



# flake8: noqa
##############################################################################
import os
from pathlib import Path

PYTORCH_MODEL_CODE = """
import torch
import torch.nn as nn

from utils.myutil import my_util_function

class MyModel(nn.Module):
    def __init__(self):
        super(MyModel, self).__init__()
        self.main = torch.nn.Sequential(
            torch.nn.Linear(3, 1),
            torch.nn.Flatten(0, 1)
        )

    def forward(self, input):
        print('Using my util function', my_util_function())
        return self.main(input)
"""

UTIL_CODE = """
def my_util_function():
    return 1
"""

PYTORCH_EG_PATH = 'pytorch_eg'
path = Path(PYTORCH_EG_PATH)
path.mkdir(parents=True, exist_ok=True)
path = Path(f'{PYTORCH_EG_PATH}/utils')
path.mkdir(parents=True, exist_ok=True)

with open(f'{PYTORCH_EG_PATH}/my_pytorch_model.py', 'w') as f:
    f.write(PYTORCH_MODEL_CODE)

with open(f'{PYTORCH_EG_PATH}/utils/myutil.py', 'w') as f:
    f.write(UTIL_CODE)

with open(f'{PYTORCH_EG_PATH}/utils/__init__.py', 'w') as f:
    f.write('')

current_dir = os.getcwd()
os.chdir(PYTORCH_EG_PATH)
from my_pytorch_model import MyModel

model = MyModel()
os.chdir(current_dir)

from truss.build import mk_truss

# You can take a whole directory
ms = mk_truss(model, model_files=['pytorch_eg/'], data_files=[], target_directory='test_pytorch')
# You can also take single files or globs
ms2 = mk_truss(model, model_files=['pytorch_eg/utils/*.py', 'pytorch_eg/my_pytorch_model.py'], data_files=[], target_directory='test_pytorch2')

ms.docker_build_string
ms.predict([[0,0,0]])


import random
import string

##############################################################################
import pandas as pd
from sklearn.datasets import load_iris
from sklearn.ensemble import RandomForestClassifier

from truss.build import scaffold

random_suffix = ''.join([random.choice(string.ascii_letters) for _ in range(5)])
rfc = RandomForestClassifier()

iris = load_iris()
feature_names = iris['feature_names']
class_labels = list(iris['target_names'])
data_x = iris['data']
data_y = iris['target']
data_x = pd.DataFrame(data_x, columns=feature_names)
rfc.fit(data_x, data_y)

ms = mk_truss(rfc, model_files=[], data_files=[])
ms.docker_build_string
ms.predict([[0,0,0,0]])


##############################################################################
EMBEDDING_REQUIREMENTS = """
tensorflow-hub==0.10.0
tensorflow==2.5.0
scikit-learn==1.0.2
"""

EMBEDDING_UTIL_CODE = """
import numpy as np


def get_top_k_indices(arr, k=5):
    return arr.argsort(axis=0)[-k::][::-1].tolist()


def create_reference_embeddings():
    random_fake_embeddings = np.random.randn(100,512)/10
    np.save('embeddings.npy', random_fake_embeddings)

"""

EMBEDDING_MODEL_CODE = """
import numpy as np
import tensorflow_hub as hub
from sklearn.metrics.pairwise import cosine_similarity
import pathlib

from utils.embedding_util import get_top_k_indices

ENCODER_URL = 'https://tfhub.dev/google/universal-sentence-encoder/4'
REFERENCE_EMBED = 'embeddings.npy'


class MyEmbeddingModel:
    def __init__(self):
        self.embed = None
        self.reference_embeddings = None

    def load(self):
        self.embed = hub.load(ENCODER_URL)
        self.reference_embeddings = np.load(pathlib.Path('model', REFERENCE_EMBED).as_posix())

    def predict(self, inputs):
        # do the prediction
        embedding = self.embed(inputs)
        # compare the embedding with a reference dataset
        embedding_cosine_similarity = cosine_similarity(self.reference_embeddings, embedding)
        # find the most similar
        top_k_similar = get_top_k_indices(embedding_cosine_similarity)
        # return raw and processed results
        return {
            "embedding": embedding.numpy().tolist(),
            "top_k_similar": top_k_similar,
        }

"""
from pathlib import Path

from truss.build import scaffold_custom

path = Path('test_folder')
path.mkdir(parents=True, exist_ok=True)
path = Path('test_folder/utils')
path.mkdir(parents=True, exist_ok=True)
# Create the model file
with open('test_folder/embedding_model.py', 'w') as f:
    f.write(EMBEDDING_MODEL_CODE)

# Create some utils file
with open('test_folder/utils/embedding_util.py', 'w') as f:
    f.write(EMBEDDING_UTIL_CODE)

with open('test_folder/utils/__init__.py', 'w') as f:
    f.write('')

# Create the requirements file
with open('test_folder/embedding_reqs.txt', 'w') as f:
    f.write(EMBEDDING_REQUIREMENTS)

# Create a fake dataset to include with the deployment
from test_folder.utils.embedding_util import create_reference_embeddings

create_reference_embeddings()

mk_truss = scaffold_custom(
    model_files=['test_folder', 'test_folder/utils/*.py', 'embeddings.npy'],
    target_directory='test_custom',
    requirements_file='test_folder/embedding_reqs.txt',
    model_class='MyEmbeddingModel'
)
mk_truss.docker_build_string
mk_truss.predict(['hello world', 'bar baz'])


##############################################################################
import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras import layers
from tensorflow.keras.layers.experimental import preprocessing

url = 'http://archive.ics.uci.edu/ml/machine-learning-databases/auto-mpg/auto-mpg.data'
column_names = ['MPG', 'Cylinders', 'Displacement', 'Horsepower', 'Weight',
                'Acceleration', 'Model Year', 'Origin']

raw_dataset = pd.read_csv(url, names=column_names,
                          na_values='?', comment='\t',
                          sep=' ', skipinitialspace=True)

dataset = raw_dataset.copy()
dataset.isna().sum()

dataset = dataset.dropna()
dataset['Origin'] = dataset['Origin'].map({1: 'USA', 2: 'Europe', 3: 'Japan'})
dataset = pd.get_dummies(dataset, columns=['Origin'], prefix='', prefix_sep='')

train_dataset = dataset.sample(frac=0.8, random_state=0)
test_dataset = dataset.drop(train_dataset.index)

train_dataset.describe().transpose()

train_features = train_dataset.copy()
test_features = test_dataset.copy()

train_labels = train_features.pop('MPG')
test_labels = test_features.pop('MPG')

normalizer = preprocessing.Normalization(axis=-1)
normalizer.adapt(np.array(train_features))

linear_model = tf.keras.Sequential([
    normalizer,
    layers.Dense(units=1)
])

linear_model.compile(
    optimizer=tf.optimizers.Adam(learning_rate=0.1),
    loss='mean_absolute_error'
)

history = linear_model.fit(
    train_features, train_labels,
    epochs=100,
    # suppress logging
    verbose=0,
    # Calculate validation results on 20% of the training data
    validation_split = 0.2
)

linear_model.predict(train_features[:10])

from truss.build import scaffold

scaff = mk_truss(model=linear_model)
scaff.docker_build_string
scaff.predict([[0, 0, 0, 0, 0, 0, 0, 0, 0]])

##############################################################################

from truss.constants import HUGGINGFACE_TRANSFORMER
from truss.definitions.base import build_scaffold_directory
from truss.definitions.huggingface_transformer import \
    HuggingFaceTransformerPipelineScaffold

built_scaffold_dir = build_scaffold_directory(
    HUGGINGFACE_TRANSFORMER
)
mk_truss = HuggingFaceTransformerPipelineScaffold(model_type='text-generation', path_to_scaffold=built_scaffold_dir)
mk_truss.predict([{'text_inputs': 'hello world'}])


##############################################################################
import os
from pathlib import Path

PYTORCH_MODEL_CODE = """
import torch
import torch.nn as nn

class MyModelWithArgs(nn.Module):
    def __init__(self, nlayers, layer_size=64):
        super(MyModelWithArgs, self).__init__()
        self.nlayers = nlayers
        self.layer_size = layer_size
        self.input_layer = nn.Linear(3, self.layer_size)
        self.dense_layers = nn.ModuleList([nn.Linear(self.layer_size, self.layer_size) for _ in range(self.nlayers)])
        self.output_layer = nn.Linear(self.layer_size, 1)

    def forward(self, input):
        x = self.input_layer(input)
        for layer in self.dense_layers:
            x = torch.nn.functional.relu(layer(x))
        x = self.output_layer(x)
        return x

"""

PYTORCH_WITH_ARGS_PATH = 'pytorch_with_args'
MODEL_INIT_ARGS = {'nlayers': 2, 'layer_size': 32}

path = Path(PYTORCH_WITH_ARGS_PATH)
path.mkdir(parents=True, exist_ok=True)

with open(f'{PYTORCH_WITH_ARGS_PATH}/my_pytorch_model.py', 'w') as f:
    f.write(PYTORCH_MODEL_CODE)


current_dir = os.getcwd()
os.chdir(PYTORCH_WITH_ARGS_PATH)
from my_pytorch_model import MyModelWithArgs

model = MyModelWithArgs(**MODEL_INIT_ARGS)
os.chdir(current_dir)

from truss.build import scaffold

ms = mk_truss(model, model_files=[PYTORCH_WITH_ARGS_PATH], data_files=[], target_directory='test_pytorch_with_args', model_init_parameters=MODEL_INIT_ARGS)
ms.predict([[0, 0, 0]])


##############################################################################
# Simple custom with args
import os
from pathlib import Path

CUSTOM_MODEL_CODE = """
class MyCustomModelWithArgs:
    def __init__(self, arg1, arg2, keyword_arg=64):
        self.arg1 = arg1
        self.arg2 = arg2
        self.keyword_arg = keyword_arg
        print(f'arg1: {self.arg1}')
        print(f'arg2: {self.arg2}')
        print(f'keyword_arg: {self.keyword_arg}')

    def load(self):
        print('loading model')
        print(f'arg1: {self.arg1}')
        print(f'arg2: {self.arg2}')
        print(f'keyword_arg: {self.keyword_arg}')

    def predict(self, input):
        return input

"""

CUSTOM_WITH_ARGS_PATH = 'custom_with_args'
MODEL_INIT_ARGS = {'arg1': 2, 'arg2': 32, 'keyword_arg': 64}

path = Path(CUSTOM_WITH_ARGS_PATH)
path.mkdir(parents=True, exist_ok=True)

with open(f'{CUSTOM_WITH_ARGS_PATH}/my_custom_model.py', 'w') as f:
    f.write(CUSTOM_MODEL_CODE)

from truss.build import scaffold_custom

ms = scaffold_custom(
    model_files=[CUSTOM_WITH_ARGS_PATH],
    target_directory='test_custom_with_args',
    model_class='MyCustomModelWithArgs',
    model_init_parameters=MODEL_INIT_ARGS,
)
ms.predict([[0, 0, 0]])


##############################################################################
# Simple custom without args
import os
from pathlib import Path

CUSTOM_MODEL_CODE = """
class MyCustomModel:
    def load(self):
        print('loading model')

    def predict(self, input):
        return input

"""

CUSTOM_WITHOUT_ARGS_PATH = 'custom_without_args'

path = Path(CUSTOM_WITHOUT_ARGS_PATH)
path.mkdir(parents=True, exist_ok=True)

with open(f'{CUSTOM_WITHOUT_ARGS_PATH}/my_custom_model.py', 'w') as f:
    f.write(CUSTOM_MODEL_CODE)

from truss.build import scaffold_custom

ms = scaffold_custom(
    model_files=[CUSTOM_WITHOUT_ARGS_PATH],
    target_directory='test_custom_without_args',
    model_class='MyCustomModel',
)
ms.predict([[0, 0, 0]])
